@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 """
 This module has been modified from the chopped-out of class from nplab_.
-It is a serial instrument class to simplify the process of interfacing with 
+It is a serial instrument class to simplify the process of interfacing with
 equipment that talks on a serial port.  The idea is that your instrument can
 subclass :class:`ExtensibleSerialInstrument` and provide methods to control the
 hardware, which will mostly consist of `self.query()` commands. It also has
@@ -27,7 +27,6 @@ from serial import STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO
 import io
 import time
 import logging
-import warnings
 
 
 class ExtensibleSerialInstrument(object):
@@ -39,14 +38,14 @@ class ExtensibleSerialInstrument(object):
     a call to the `query` method.  This writes a message and returns the reply.
     This has been hacked together from the nplab_ MessageBusInstrument and SerialInstrument
     classes.
-    
+
     **Threading Notes**
-    
+
     The message bus protocol includes a property, `communications_lock`.  All
     commands that use the communications bus should be protected by this lock.
-    It's also permissible to use it to protect sequences of calls to the bus 
+    It's also permissible to use it to protect sequences of calls to the bus
     that must be atomic (e.g. a multi-part exchange of messages).  However, try
-    not to hold it too long - or odd things might happen if other threads are 
+    not to hold it too long - or odd things might happen if other threads are
     blocked for a long time.  The lock is reentrant so there's no issue with
     acquiring it twice.
     """
@@ -73,29 +72,32 @@ class ExtensibleSerialInstrument(object):
 
     def open(self, port=None, quiet=True):
         """Open communications with the serial port.
-        
+
         If no port is specified, it will attempt to autodetect.  If quiet=True
         then we don't warn when ports are opened multiple times.
         """
         with self.communications_lock:
-            if hasattr(self, "_ser") and self._ser.isOpen():
-                if not quiet:
-                    logging.warning("Attempted to open an already-open port!")
-                return
+            if hasattr(self, "_ser"):
+                if self._ser.isOpen(): # pylint: disable=access-member-before-definition
+                    if not quiet:
+                        logging.warning("Attempted to open an already-open port!")
+                    return
             if port is None:
                 port = self.find_port()
-            assert (
-                port is not None
-            ), "We don't have a serial port to open, meaning you didn't specify a valid port.  Are you sure the instrument is connected?"
+            if port is None:
+                raise FileNotFoundError("We don't have a serial port to open, meaning you "
+                                        "didn't specify a valid port.  Are you sure the "
+                                        "instrument is connected?")
             logging.info("Creating serial.Serial instance...")
-            self._ser = serial.Serial(port, **self.port_settings)
+            if isinstance(port, DummySerialDevice):
+                self._ser = port
+            else:
+                self._ser = serial.Serial(port, **self.port_settings)
             logging.info(f"Created {self._ser}")
-            # the block above wraps the serial IO layer with a text IO layer
-            # this allows us to read/write in neat lines.  NB the buffer size must
-            # be set to 1 byte for maximum responsiveness.
-            assert (
-                self.test_communications()
-            ), "The instrument doesn't seem to be responding.  Did you specify the right port?"
+
+            if not self.test_communications():
+                raise IOError("The instrument doesn't seem to be responding.  Did you specify"
+                              " the right port?")
 
     def close(self):
         """Cleanly close the device. Includes proper logging statements."""
@@ -287,15 +289,14 @@ class ExtensibleSerialInstrument(object):
                 reply = self.readline().strip()
                 res = re.search(response_regex, reply, flags=re_flags)
                 if res is not None:
-                    warnings.warn(
+                    logging.warning(
                         "Query suceeded after initially receieving unmatched response ('%s') to '%s'. Match pattern /%s/ (generated regex /%s/)"
                         % (
                             original_reply,
                             query_string,
                             response_string,
                             response_regex,
-                        ),
-                        RuntimeWarning,
+                        )
                     )
             else:
                 raise ValueError(
@@ -327,7 +328,7 @@ class ExtensibleSerialInstrument(object):
         return self.parsed_query(query_string, "%f", **kwargs)
 
     def test_communications(self):
-        """Check if the device is available on the current port.  
+        """Check if the device is available on the current port.
         
         This should be overridden by subclasses.  Assume the port has been
         successfully opened and the settings are as defined by self.port_settings.
@@ -516,3 +517,142 @@ class QueriedProperty(object):
         if self.fdel is None:
             raise AttributeError("can't delete attribute")
         self.fdel(obj)
+
+class DummySerialDevice():
+    """This class exists to allow unit testing without serial port hardware.
+
+    It can be passed to an ``ExtensibleSerialInstrument`` in place of the
+    serial port name, and will emulate a device using stored responses.
+
+    It works on a line-by-line basis, so it will only do anything once
+    it gets a termination character (by default, newline).  It tries to give
+    the same interface as ``pyserial.Serial`` at least as far as the 
+    methods used by ``ExtensibleSerialInstrument`` go.  
+
+    The ``timeout`` property will be used to cause delays if there is
+    no response to give - but infinite timeouts raise an exception rather
+    than attempt to do anything clever with threads.
+
+    You can set ``print_buffers`` to True in order to make it print the
+    input and output buffers each time the "port" is written to or read
+    from.
+    """
+    termination_character = "\n"
+    responses = ()
+    unmatched_response = "error!\n"
+    timeout = 0
+    print_buffers = False
+
+    read_buffer = b""
+    write_buffer = b""
+
+    def __init__(self):
+        pass
+
+    def isOpen(self):
+        return True
+
+    def write(self, bytes):  
+        """Accept data being written to the port
+        
+        This is data being sent out of the serial port, and is
+        stored in the ``write buffer``.
+        """
+        self.write_buffer += bytes
+        if self.print_buffers:
+            print("Write buffer: [[{}]]".format(self.write_buffer))
+        # Process messages once we've received a whole line
+        index = self.write_buffer.find(self.termination_character.encode())
+        if index > 0:
+            self.respond_to(self.write_buffer[:index].decode("utf8"))
+            self.write_buffer = self.write_buffer[index + 1:]
+
+    def inWaiting(self):
+        """Return the number of bytes in the input buffer
+
+        NB the input buffer is for bytes being read from the port
+        """
+        return len(self.read_buffer)
+
+    def flushInput(self):
+        """Flush the input buffer, so there are no previously-arrived
+        bytes when we read a response."""
+        self.read_buffer = b""
+
+    def register_response(self, regular_expression, response_function):
+        """Match a query and respond to it with a function.
+
+        See ``respond_to`` for details of how this works.
+        """
+        if isinstance(response_function, str):
+            response_str = response_function
+            response_function = lambda groups: response_str
+        self.responses += ((regular_expression, response_function),)
+    
+    def remove_response(self, regex_to_remove):
+        """Remove a previously registered response
+
+        The regular expression string must be exactly the same as the
+        one you are trying to remove.
+        """
+        self.responses = tuple(
+            (regex, response) for (regex, response) in self.responses 
+                              if regex != regex_to_remove)
+
+    def replace_response(self, regex, response_function):
+        """Remove a response and replace it with another.
+
+        Order is not preserved.
+        """
+        self.remove_response(regex)
+        self.register_response(regex, response_function)
+
+    def respond_to(self, message):
+        """Do something in response to a message.
+
+        Each time we receive a termination character, we attempt to match
+        the message (which doesn't include the termination character) with
+        the regular expressions in the ``responses`` property.  The first
+        one that matches is used.
+
+        The regular expression match object's ``groups()`` value is passed
+        to the function, which can then use it to return its response.
+        """
+        for regular_expression, response_function in self.responses:
+            match = re.match(regular_expression, message)
+            if match:
+                self.read_buffer += response_function(match.groups()).encode()
+                return
+        self.read_buffer += self.unmatched_response.encode()
+
+    def readline(self):
+        """Emulate reading a line from the serial port."""
+        if self.print_buffers:
+            print("Read buffer [[{}]]".format(self.read_buffer))
+        index = self.read_buffer.find(self.termination_character.encode())
+        if index > 0:
+            line = self.read_buffer[:index + 1]
+            self.read_buffer = self.read_buffer[index + 1:]
+            return line
+        else:
+            if self.timeout:
+                time.sleep(self.timeout)
+            else:
+                raise IOError("readline was called with no timeout, and there's nothing to read.")
+    
+    def read(self, size=1):
+        """Emulate reading some bytes from the serial port."""
+        if len(self.read_buffer) < size:
+            chunk = self.read_buffer[:size]
+            self.read_buffer = self.read_buffer[size:]
+            return chunk
+        else:
+            raise IOError("Currently this dummy class can't cope with timeouts properly")
+
+    def open(self):
+        """Open the serial port (currently does nothing)"""
+        pass
+
+    def close(self):
+        """Close the serial port (currently does nothing)"""
+        pass
